"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.cli = void 0;
const utils_1 = require("./utils");
const constants_1 = require("./constants");
const evmAtomicTx_1 = require("./evmAtomicTx");
const pvmAtomicTx_1 = require("./pvmAtomicTx");
const addValidator_1 = require("./addValidator");
const dist_1 = require("@flarenetwork/flarejs/dist");
const logging_1 = __importDefault(require("logging"));
const addDelegator_1 = require("./addDelegator");
const logger = (0, logging_1.default)('info');
function cli(program) {
    return __awaiter(this, void 0, void 0, function* () {
        // global configurations
        program
            .option("--network <network>", "Network name (flare or costwo)", 'flare')
            .option("--env-path <path>", "Path to the .env file", 'env');
        // information about the network
        program
            .command("info").description("Relevant information")
            .argument("<type>", "Type of information")
            .action((type) => __awaiter(this, void 0, void 0, function* () {
            const options = program.opts();
            const ctx = (0, constants_1.contextEnv)(options.envPath, options.network);
            if (type == 'addresses') {
                getAddressInfo(ctx);
            }
            else if (type == 'balance') {
                yield getBalanceInfo(ctx);
            }
            else if (type == 'network') {
                getNetworkInfo(ctx);
            }
            else if (type == 'livenetwork') {
                // implement this nicely
            }
            else if (type == 'validators') {
                yield getValidatorInfo(ctx);
            }
        }));
        // moving funds from one chain to another
        program
            .command("crosschain").description("Move funds from one chain to another")
            .argument("<type>", "Type of a crosschain transaction")
            .option("-a, --amount <amount>", "Amount to transfer")
            .option("-f, --fee <fee>", "Fee of a transaction")
            .action((type, options) => __awaiter(this, void 0, void 0, function* () {
            options = Object.assign(Object.assign({}, options), program.opts());
            const ctx = (0, constants_1.contextEnv)(options.envPath, options.network);
            if (type == 'exportCP') {
                yield exportCP(ctx, options.amount, options.fee);
            }
            else if (type == 'importCP') {
                yield importCP(ctx);
            }
            else if (type == 'exportPC') {
                yield exportPC(ctx, options.amount);
            }
            else if (type == 'importPC') {
                yield importPC(ctx, options.fee);
            }
        }));
        // staking
        program
            .command("stake").description("Stake funds on the P-chain")
            .option("-n, --node-id <nodeID>", "The staking node's id")
            .option("-a, --amount <amount>", "Amount to stake")
            .option("-s, --start-time <start-time>", "Start time of the staking process")
            .option("-e, --end-time <end-time>", "End time of the staking process")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = Object.assign(Object.assign({}, options), program.opts());
            const ctx = (0, constants_1.contextEnv)(options.envPath, options.network);
            yield stake(ctx, options.nodeId, options.amount, options.startTime.startsWith("now+") ? (0, utils_1.parseRelativeTime)(options.startTime) : options.startTime, options.endTime.startsWith("now+") ? (0, utils_1.parseRelativeTime)(options.endTime) : options.endTime);
        }));
        // delegating
        program
            .command("delegate").description("Delegate funds on the P-chain")
            .option("-n, --node-id <nodeID>", "The staking node's id")
            .option("-a, --amount <amount>", "Amount to delegate")
            .option("-s, --start-time <start-time>", "Start time of the delegation process")
            .option("-e, --end-time <end-time>", "End time of the delegation process")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = Object.assign(Object.assign({}, options), program.opts());
            const ctx = (0, constants_1.contextEnv)(options.envPath, options.network);
            yield delegate(ctx, options.nodeId, options.amount, options.startTime.startsWith("now+") ? (0, utils_1.parseRelativeTime)(options.startTime) : options.startTime, options.endTime.startsWith("now+") ? (0, utils_1.parseRelativeTime)(options.endTime) : options.endTime);
        }));
        // converting between address formats
        program
            .command("convert").description("Utility for conversion of address formats")
            .argument("<type>", "Type of conversion")
            .option("-p, --public-key <pubk>", "User's secp256k1 public key (encoded either 0x02, 0x03 or 0x04)")
            .action((type, options) => {
            options = Object.assign(Object.assign({}, options), program.opts());
            const ctx = (0, constants_1.contextEnv)(options.envPath, options.network);
            if (type == 'public-key-to-p-chain-address') {
                getAddressFromPublicKey(ctx, options.publicKey);
            }
        });
    });
}
exports.cli = cli;
function getAddressInfo(ctx) {
    const [pubX, pubY] = (0, utils_1.privateKeyToPublicKey)(Buffer.from(ctx.privkHex, 'hex'));
    const compressedPubKey = (0, utils_1.compressPublicKey)(pubX, pubY).toString('hex');
    logger.info(`X-chain address: ${ctx.xAddressBech32}`);
    logger.info(`P-chain address: ${ctx.pAddressBech32}`);
    logger.info(`C-chain address hex: ${ctx.cAddressHex}`);
    logger.info(`secp256k1 public key: 0x${compressedPubKey}`);
}
function getBalanceInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        let cbalance = (new dist_1.BN(yield ctx.web3.eth.getBalance(ctx.cAddressHex))).toString();
        let pbalance = (new dist_1.BN((yield ctx.pchain.getBalance(ctx.pAddressBech32)).balance)).toString();
        let xbalance = (new dist_1.BN((yield ctx.xchain.getBalance(ctx.xAddressBech32, ctx.avaxAssetID)).balance)).toString();
        cbalance = (0, utils_1.integerToDecimal)(cbalance, 18);
        pbalance = (0, utils_1.integerToDecimal)(pbalance, 9);
        xbalance = (0, utils_1.integerToDecimal)(xbalance, 9);
        logger.info(`${ctx.cAddressHex}: ${cbalance}`);
        logger.info(`${ctx.pAddressBech32}: ${pbalance}`);
        logger.info(`${ctx.xAddressBech32}: ${xbalance}`);
    });
}
function getNetworkInfo(ctx) {
    const pchainId = ctx.pchain.getBlockchainID();
    const cchainId = ctx.cchain.getBlockchainID();
    const xchainId = ctx.xchain.getBlockchainID();
    logger.info(`blockchainId for P-chain: ${pchainId}`);
    logger.info(`blockchainId for C-chain: ${cchainId}`);
    logger.info(`blockchainId for X-chain: ${xchainId}`);
    logger.info(`assetId: ${ctx.avaxAssetID}`);
}
function getValidatorInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const pending = yield ctx.pchain.getPendingValidators();
        const current = yield ctx.pchain.getCurrentValidators();
        const fpending = JSON.stringify(pending, null, 2);
        const fcurrent = JSON.stringify(current, null, 2);
        logger.info(`pending: ${fpending}`);
        logger.info(`current: ${fcurrent}`);
    });
}
function exportCP(ctx, amount, fee) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = new dist_1.BN((0, utils_1.decimalToInteger)(amount, 9));
        const ffee = (fee === undefined) ?
            fee : new dist_1.BN((0, utils_1.decimalToInteger)(fee, 9));
        const { txid, usedFee } = yield (0, evmAtomicTx_1.exportTxCP)(ctx, famount, ffee);
        if (fee !== usedFee)
            logger.info(`Used fee of ${usedFee}`);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function importCP(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const { txid } = yield (0, pvmAtomicTx_1.importTxCP)(ctx);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function exportPC(ctx, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = (amount === undefined) ?
            amount : new dist_1.BN((0, utils_1.decimalToInteger)(amount, 9));
        const { txid } = yield (0, pvmAtomicTx_1.exportTxPC)(ctx, famount);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function importPC(ctx, fee) {
    return __awaiter(this, void 0, void 0, function* () {
        const ffee = (fee === undefined) ?
            fee : new dist_1.BN((0, utils_1.decimalToInteger)(fee, 9));
        const { txid, usedFee } = yield (0, evmAtomicTx_1.importTxPC)(ctx, ffee);
        if (fee !== usedFee)
            logger.info(`Used fee of ${usedFee}`);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function stake(ctx, nodeID, amount, start, end) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = new dist_1.BN((0, utils_1.decimalToInteger)(amount, 9));
        const { txid } = yield (0, addValidator_1.addValidator)(ctx, nodeID, famount, new dist_1.BN(start), new dist_1.BN(end));
        logger.info(`Success! TXID: ${txid}`);
    });
}
function delegate(ctx, nodeID, amount, start, end) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = new dist_1.BN((0, utils_1.decimalToInteger)(amount, 9));
        const { txid } = yield (0, addDelegator_1.addDelegator)(ctx, nodeID, famount, new dist_1.BN(start), new dist_1.BN(end));
        logger.info(`Success! TXID: ${txid}`);
    });
}
function getAddressFromPublicKey(ctx, pubk) {
    const address = (0, utils_1.publicKeyToBech32AddressString)(ctx.config.hrp, pubk);
    logger.info(`P-chain address: P-${address}`);
}
